/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    view_selection::ArmarXObjects::gaze_control
 * @author     Johann Mantel ( j-mantel at gmx dot net )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "Component.h"

#include <cstdlib>

#include <SimoxUtility/algorithm/string/string_tools.h>
#include <SimoxUtility/color/Color.h>

#include <ArmarXCore/core/ManagedIceObjectPlugin.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/time/ice_conversions.h>
#include <ArmarXCore/libraries/DecoupledSingleComponent/Decoupled.h>

#include <RobotAPI/libraries/armem/client/MemoryNameSystem.h>
#include <RobotAPI/libraries/armem/client/Writer.h>
#include <RobotAPI/libraries/armem/client/query.h>
#include <RobotAPI/libraries/armem/core/wm/ice_conversions.h>
#include <RobotAPI/libraries/aron/common/aron_conversions/core.h>

#include <armarx/view_selection/common/controller_types.h>
#include <armarx/view_selection/gaze_controller/aron/ControllerConfigBase.aron.generated.h>
#include <armarx/view_selection/gaze_controller/aron_conversions.h>
#include <armarx/view_selection/gaze_controller/atan2/controller_description.h>
#include <armarx/view_selection/gaze_controller/controller_descriptions.h>
#include <armarx/view_selection/gaze_controller/gaze_ik/controller_description.h>
#include <armarx/view_selection/gaze_controller/hemisphere/controller_description.h>
#include <armarx/view_selection/gaze_scheduler/ControllerHandler.h>
#include <armarx/view_selection/gaze_scheduler/ControllerHandlerInterface.h>
#include <armarx/view_selection/gaze_scheduler/Scheduler.h>
#include <armarx/view_selection/gaze_targets.h>
#include <armarx/view_selection/gaze_targets/AttentionType.h>
#include <armarx/view_selection/gaze_targets/TargetPriority.h>
#include <armarx/view_selection/memory/constants.h>

namespace armarx::view_selection::gaze_scheduler
{

    const std::string Component::defaultName = memory::constants::GazeSchedulerProviderName;

    armarx::PropertyDefinitionsPtr
    Component::createPropertyDefinitions()
    {
        armarx::PropertyDefinitionsPtr def =
            new armarx::ComponentPropertyDefinitions(getConfigIdentifier());
        def->optional(properties.controllerType,
                      "controllerType",
                      "The type of NJointController to use, e.g. GazeController")
            .map("GazeIK", common::ControllerType::GazeControllerGazeIK)
            .map("Atan2", common::ControllerType::GazeControllerAtan2)
            .map("Hemisphere", common::ControllerType::GazeControllerHemisphere);
        // def->optional(properties.controllerName,
        //               "controllerName",
        //               "The name under which the controller is registered");
        def->optional(properties.defaultTarget,
                      "properties.defaultTarget",
                      "Local coordinates to Position where the robot looks in idle phases.");

        def->optional(properties.targetVizAlpha, "properties.targetVizAlpha", "");
        def->optional(properties.visuEnabled, "properties.visuEnabled", "");
        def->optional(properties.dummy, "properties.dummy", "");

        def->optional(properties.rtUnitDynamicallyLoadLibraries,
                      "p.rtUnitDynamicallyLoadLibraries",
                      "If enabled, the controller library will automatically be loaded in the "
                      "RobotUnit. If disabled (desired), you have to make sure that the controller "
                      "library is listed in the RobotUnit's property `LoadLibraries`");

        return def;
    }

    void
    Component::onMemoryUpdated(const std::vector<armarx::armem::MemoryID>& updatedSnapshotIDs)
    {
        if (properties.dummy)
        {
            // THE COMPONENT IS SET TO DISABLED
            return;
        }

        ARMARX_CHECK(scheduler.has_value());

        namespace armem = armarx::armem;
        namespace constants = ::armarx::view_selection::memory::constants;
        const armem::MemoryID coreSegmentID(constants::ViewMemoryName,
                                            constants::GazeTargetCoreSegmentName);
        const armem::MemoryID schedulerProviderID(constants::ViewMemoryName,
                                                  constants::GazeTargetCoreSegmentName,
                                                  constants::GazeSchedulerProviderName);
        armem::client::Reader reader = memoryNameSystem().getReader(coreSegmentID);

        ARMARX_VERBOSE << "Received memory update: " << updatedSnapshotIDs;

        const auto result = reader.queryMemoryIDs(updatedSnapshotIDs);
        if (result.success)
        {
            const armem::wm::Memory& memory = result.memory;
            for (const auto& memId : updatedSnapshotIDs)
            {
                if (not armem::contains(schedulerProviderID, memId))
                {
                    // foreign memory updates are targetRequests
                    ARMARX_VERBOSE << "new Request received: " << memId.str();
                    const auto* snapshot = memory.findLatestInstance(memId);

                    if (snapshot != nullptr)
                    {
                        gaze_targets::GazeTarget gazeTarget =
                            armarx::aron::fromAron<gaze_targets::GazeTarget>(
                                gaze_targets::arondto::GazeTarget::FromAron(snapshot->data()));

                        scheduler->submitToQueue(gazeTarget);
                    }
                    else
                    {
                        ARMARX_VERBOSE << "No memory found with id:" << memId;
                    }
                }
            }

            scheduler->scheduleNextTarget();
        }
        else
        {
            ARMARX_VERBOSE << "Query failed: " << result.errorMessage;
        }
    }

    void
    Component::onInitComponent()
    {
        // Topics and properties defined above are automagically registered.
    }

    void
    Component::onConnectComponent()
    {

        // Load the controllers that we are using here.
        if (properties.rtUnitDynamicallyLoadLibraries)
        {
            ARMARX_WARNING
                << "Dynamic loading of the controller library is deactivated (and this "
                   "is also desired). If loading the controllers is failing subsequently, "
                   "make sure that the RobotUnit's property `LoadLibraries` contains "
                   "`armarx_view_selection:armarx_armarx_view_selection_gaze_controller_{"
                   "gaze_ik,atan2,hemisphere}`.";
        }


        // obtain info about robot
        const std::string robotName = getControlComponentPlugin()
                                          .getRobotUnitPlugin()
                                          .getRobotUnit()
                                          ->getKinematicUnit()
                                          ->getRobotName();


        controllerHandler = [&]() -> std::unique_ptr<ControllerHandlerInterface>
        {
            const ControllerHandlerInjectedServices srv{
                .controlComponentPlugin =
                    std::experimental::make_observer(&getControlComponentPlugin())};

            const ControllerHandlerParams params{.robotName = robotName};

            switch (properties.controllerType)
            {
                case common::ControllerType::GazeControllerGazeIK:
                    return std::make_unique<
                        ControllerHandler<common::ControllerType::GazeControllerGazeIK>>(srv,
                                                                                         params);
                case common::ControllerType::GazeControllerAtan2:
                    return std::make_unique<
                        ControllerHandler<common::ControllerType::GazeControllerAtan2>>(srv,
                                                                                        params);
                case common::ControllerType::GazeControllerHemisphere:
                    return std::make_unique<
                        ControllerHandler<common::ControllerType::GazeControllerHemisphere>>(
                        srv, params);
            }

            return nullptr;
        }();

        {
            const Scheduler::InjectedServices srv{
                .virtualRobotReader =
                    std::experimental::make_observer(&virtualRobotReaderPlugin->get()),
                .arviz = std::experimental::make_observer(&arviz),
                .controllerHandler = std::experimental::make_observer(controllerHandler.get())};


            const auto elements = simox::alg::split(properties.defaultTarget, ",");
            ARMARX_CHECK_EQUAL(elements.size(), 3);

            const Eigen::Vector3f defaultTarget{
                std::stof(elements.at(0)), std::stof(elements.at(1)), std::stof(elements.at(2))};

            const Scheduler::Params params{.robotName = robotName,
                                           .defaultTarget = defaultTarget,
                                           .visuEnabled = properties.visuEnabled,
                                           .targetVizAlpha = properties.targetVizAlpha};

            scheduler.emplace(srv, params);
        }


        // Do things after connecting to topics and components.
        ARMARX_INFO << "Connecting controller.";

        // robot = virtualRobotReaderPlugin->get().getRobotWaiting(
        //     robotName,
        //     armarx::DateTime::Invalid(),
        //     VirtualRobot::RobotIO::RobotDescription::eStructure);
        // ARMARX_CHECK_NOT_NULL(robot);


        // subscribe to memory updates
        {
            namespace constants = ::armarx::view_selection::memory::constants;
            const armem::MemoryID coreSegmentID(constants::ViewMemoryName,
                                                constants::GazeTargetCoreSegmentName);

            // FIXME use handle to unsubscribe
            ARMARX_INFO << "Subscribing to memory";
            memoryNameSystem().subscribe(coreSegmentID,
                                         [&](const std::vector<armem::MemoryID>& updatedSnapshotIDs)
                                         { onMemoryUpdated(updatedSnapshotIDs); });
            ARMARX_INFO << "Done subscribing to memory";
        }


        // task = new armarx::SimplePeriodicTask<>([&]() { runPeriodically(); }, 100);
        // task->start();
    }

    void
    Component::onDisconnectComponent()
    {
        // ARMARX_VERBOSE << "Disconnecting controller.";
        scheduler.reset();
        controllerHandler.reset(); // FIXME needed?
    }

    void
    Component::onExitComponent()
    {
        // delete the controller wrapper => controller will be deactivated and deleted
        scheduler.reset();
    }

    std::string
    Component::getDefaultName() const
    {
        return Component::defaultName;
    }

    std::string
    Component::GetDefaultName()
    {
        return Component::defaultName;
    }

    void
    Component::resetPriorityQueue(const Ice::Current& c)
    {
        ARMARX_INFO << "Resetting priority queue";
        ARMARX_CHECK(scheduler.has_value());

        scheduler->resetPriorityQueue();
    }

    ARMARX_REGISTER_COMPONENT_EXECUTABLE(Component, Component::GetDefaultName());

} // namespace armarx::view_selection::gaze_scheduler
