/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    view_selection::ArmarXObjects::handover_target_provider
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @date       2022
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */


#include "Component.h"

#include <ArmarXCore/core/Component.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/time/Duration.h>
#include <ArmarXCore/libraries/DecoupledSingleComponent/Decoupled.h>

#include <armarx/view_selection/client/ViewSelection.h>
#include <armarx/view_selection/components/handover_target_provider/ComponentInterface.h>
#include <armarx/view_selection/target_provider/handover/RobotGiver.h>
#include <armarx/view_selection/target_provider/handover/RobotReceiver.h>
#include <armarx/view_selection/target_provider/handover/util.h>

#include <VisionX/libraries/armem_human/client/HumanPoseReader.h>

// Include headers you only need in function definitions in the .cpp.

// #include <Eigen/Core>

// #include <SimoxUtility/color/Color.h>


namespace armarx::view_selection::components::handover_target_provider
{

    const std::string Component::defaultName = "handover_target_provider";

    Component::Component() : viewSelection_(memoryNameSystem(), getName())
    {
        addPlugin(virtualRobotReaderPlugin_);
        addPlugin(humanPoseReaderPlugin_);
    }

    armarx::PropertyDefinitionsPtr
    Component::createPropertyDefinitions()
    {
        armarx::PropertyDefinitionsPtr def =
            new armarx::ComponentPropertyDefinitions(getConfigIdentifier());

        // Publish to a topic (passing the TopicListenerPrx).
        // def->topic(myTopicListener);

        // Subscribe to a topic (passing the topic name).
        // def->topic<PlatformUnitListener>("MyTopic");

        // Use (and depend on) another component (passing the ComponentInterfacePrx).
        // def->component(myComponentProxy)


        // Add a required property. (The component won't start without a value being set.)
        // def->required(properties.boxLayerName, "p.box.LayerName", "Name of the box layer in ArViz.");

        // Add an optionalproperty.

        def->optional(
            properties_.humanPoseProviderName, "p.humanProviderName", "OpenNIPointCloudProvider");
        // def->optional(properties_.objectPoseProviderDynamic, "p.objectProviderDynamic", "");
        def->optional(properties_.trackTime,
                      "p.trackTime",
                      "Roughly the expected average tracking time of an object in seconds.");
        def->optional(properties_.robotName, "p.robotName", "");

        return def;
    }

    void
    Component::onInitComponent()
    {
        // Topics and properties defined above are automagically registered.

        // Keep debug observer data until calling `sendDebugObserverBatch()`.
        // (Requies the armarx::DebugObserverComponentPluginUser.)
        // setDebugObserverBatchModeEnabled(true);
    }

    void
    Component::onConnectComponent()
    {
        // Do things after connecting to topics and components.

        /* (Requies the armarx::DebugObserverComponentPluginUser.)
        // Use the debug observer to log data over time.
        // The data can be viewed in the ObserverView and the LivePlotter.
        // (Before starting any threads, we don't need to lock mutexes.)
        {
            setDebugObserverDatafield("numBoxes", properties.numBoxes);
            setDebugObserverDatafield("boxLayerName", properties.boxLayerName);
            sendDebugObserverBatch();
        }
        */

        /* (Requires the armarx::ArVizComponentPluginUser.)
        // Draw boxes in ArViz.
        // (Before starting any threads, we don't need to lock mutexes.)
        drawBoxes(properties, arviz);
        */

        /* (Requires the armarx::LightweightRemoteGuiComponentPluginUser.)
        // Setup the remote GUI.
        {
            createRemoteGuiTab();
            RemoteGui_startRunningTask();
        }
        */

        robot_ = virtualRobotReaderPlugin_->get().getRobotWaiting(properties_.robotName);
        viewSelection_.connect();

        ARMARX_INFO << "Starting ...";
        start("foo", RobotRole::Receiver);
    }

    void
    Component::onDisconnectComponent()
    {
    }

    void
    Component::onExitComponent()
    {
    }

    std::string
    Component::getDefaultName() const
    {
        return Component::defaultName;
    }

    std::string
    Component::GetDefaultName()
    {
        return Component::defaultName;
    }

    void
    Component::start(
        const ::std::string& humanTrackingId,
        ::armarx::view_selection::components::handover_target_provider::RobotRole robotRole,
        const ::Ice::Current&)
    {
        if (task)
        {
            ARMARX_INFO << "Already tracked human. Will restart";
            stop();
        }

        switch (robotRole)
        {
            case Receiver:
                ARMARX_INFO << "Robot acting as receiver";
                handoverTargetProvider_ =
                    std::make_unique<target_provider::handover::RobotReceiver>(humanTrackingId,
                                                                               robot_);
                break;
            case Giver:
                ARMARX_INFO << "Robot acting as giver";
                handoverTargetProvider_ = std::make_unique<target_provider::handover::RobotGiver>(
                    humanTrackingId, robot_);
                break;
        }

        ARMARX_VERBOSE << "Starting task";
        task = new armarx::SimplePeriodicTask<>([&]() { updateTargets(); },
                                                properties_.updatePeriodMs);
        task->start();
    }

    void
    Component::stop(const ::Ice::Current&)
    {
        task->stop();
        task = nullptr;
    }

    void
    Component::updateTargets()
    {
        if (handoverTargetProvider_ == nullptr)
        {
            return;
        }

        const core::time::DateTime timestamp = armarx::Clock::Now();

        ARMARX_CHECK_NOT_NULL(robot_);
        ARMARX_CHECK(virtualRobotReaderPlugin_->get().synchronizeRobot(*robot_, timestamp));

        const armem::human::client::Reader::Query query{.providerName =
                                                            properties_.humanPoseProviderName,
                                                        .timestamp = timestamp,
                                                        .maxAge = Duration::MilliSeconds(1500)};

        const armem::human::client::Reader::Result result =
            humanPoseReaderPlugin_->get().query(query);
        if (not result)
        {
            ARMARX_VERBOSE << "Failed to query humans";
            return;
        }

        const std::vector<armem::human::HumanPose>& humans = result.humanPoses;

        if (humans.empty())
        {
            ARMARX_INFO << deactivateSpam(1) << "No humans";
            return;
        }

        const armem::human::HumanPose& engangedHuman = humans.back();

        ARMARX_VERBOSE << deactivateSpam(1) << VAROUT(engangedHuman.humanTrackingId.value());


        std::optional<std::string> headKpName =
            target_provider::handover::util::headKeypoint(engangedHuman.poseModelId);

        if (headKpName.has_value())
        {
            auto l = arviz.layer("human");
            auto headKp = engangedHuman.keypoints.at(headKpName.value());
            l.add(viz::Sphere("head").position(headKp.positionGlobal->toEigen()));
            arviz.commit(l);
        }


        const auto targets = handoverTargetProvider_->updateTargets(engangedHuman);
        if (targets.empty())
        {
            ARMARX_VERBOSE << "No targets.";
            return;
        }

        viewSelection_.commitGazeTargets(targets);
    }

    /* (Requires the armarx::LightweightRemoteGuiComponentPluginUser.)
    void
    Component::createRemoteGuiTab()
    {
        using namespace armarx::RemoteGui::Client;

        // Setup the widgets.

        tab.boxLayerName.setValue(properties.boxLayerName);

        tab.numBoxes.setValue(properties.numBoxes);
        tab.numBoxes.setRange(0, 100);

        tab.drawBoxes.setLabel("Draw Boxes");

        // Setup the layout.

        GridLayout grid;
        int row = 0;
        {
            grid.add(Label("Box Layer"), {row, 0}).add(tab.boxLayerName, {row, 1});
            ++row;

            grid.add(Label("Num Boxes"), {row, 0}).add(tab.numBoxes, {row, 1});
            ++row;

            grid.add(tab.drawBoxes, {row, 0}, {2, 1});
            ++row;
        }

        VBoxLayout root = {grid, VSpacer()};
        RemoteGui_createTab(getName(), root, &tab);
    }


    void
    Component::RemoteGui_update()
    {
        if (tab.boxLayerName.hasValueChanged() || tab.numBoxes.hasValueChanged())
        {
            std::scoped_lock lock(propertiesMutex);
            properties.boxLayerName = tab.boxLayerName.getValue();
            properties.numBoxes = tab.numBoxes.getValue();

            {
                setDebugObserverDatafield("numBoxes", properties.numBoxes);
                setDebugObserverDatafield("boxLayerName", properties.boxLayerName);
                sendDebugObserverBatch();
            }
        }
        if (tab.drawBoxes.wasClicked())
        {
            // Lock shared variables in methods running in seperate threads
            // and pass them to functions. This way, the called functions do
            // not need to think about locking.
            std::scoped_lock lock(propertiesMutex, arvizMutex);
            drawBoxes(properties, arviz);
        }
    }
    */


    /* (Requires the armarx::ArVizComponentPluginUser.)
    void
    Component::drawBoxes(const Component::Properties& p, viz::Client& arviz)
    {
        // Draw something in ArViz (requires the armarx::ArVizComponentPluginUser.
        // See the ArVizExample in RobotAPI for more examples.

        viz::Layer layer = arviz.layer(p.boxLayerName);
        for (int i = 0; i < p.numBoxes; ++i)
        {
            layer.add(viz::Box("box_" + std::to_string(i))
                      .position(Eigen::Vector3f(i * 100, 0, 0))
                      .size(20).color(simox::Color::blue()));
        }
        arviz.commit(layer);
    }
    */


    ARMARX_REGISTER_COMPONENT_EXECUTABLE(Component, Component::GetDefaultName());

} // namespace armarx::view_selection::components::handover_target_provider
