/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    view_selection::ArmarXObjects::handover_target_provider
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @date       2022
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */


#pragma once


#include <ArmarXCore/core/Component.h>
#include <ArmarXCore/core/services/tasks/TaskUtil.h>

#include <RobotAPI/libraries/ArmarXObjects/plugins/ObjectPoseClientPlugin.h>
#include <RobotAPI/libraries/RobotAPIComponentPlugins/ArVizComponentPlugin.h>
#include <RobotAPI/libraries/armem/client/plugins.h>
#include <RobotAPI/libraries/armem/client/plugins/PluginUser.h>
#include <RobotAPI/libraries/armem/client/plugins/ReaderWriterPlugin.h>
#include <RobotAPI/libraries/armem_robot_state/client/common/VirtualRobotReader.h>

#include <armarx/view_selection/client/ViewSelection.h>
#include <armarx/view_selection/components/handover_target_provider/ComponentInterface.h>
#include <armarx/view_selection/components/person_target_provider/ComponentInterface.h>
#include <armarx/view_selection/components/person_target_provider/Features.h>
#include <armarx/view_selection/target_provider/handover/HandoverTargetProvider.h>

#include <VisionX/libraries/armem_human/client/HumanPoseReader.h>

namespace armarx::view_selection::components::handover_target_provider
{

    /**
     * @brief A component to control the robot's gaze during a handover task. The behavior is as follows and justified by the papers mentioned below:
     * 
     * The receiver will look at the giver's hand during reach and transfer phase. Once it received the object, it will look at the giver's face.
     * The giver will first look at the receiver's face during reach. At a certain distance to the receiver, it will then look at the receiver's hand.
     * Once the object is exchanged, the giver will then look at the receivers face.
     * 
     * See:
     *    [1] A. Kshirsagar, M. Lim, S. Christian and G. Hoffman, "Robot Gaze Behaviors in Human-to-Robot Handovers," 
     *        in IEEE Robotics and Automation Letters, vol. 5, no. 4, pp. 6552-6558, Oct. 2020, doi: 10.1109/LRA.2020.3015692.
     * 
     *    [2] Faibish, Tair, et al. "Human preferences for robot eye gaze in human-to-robot handovers." 
     *        International Journal of Social Robotics (2022): 1-18.
     * 
     * In [1], the reach phase is analyzed. In [2], the phases reach, transfer and retreat are distinguished.
     * 
     * [1] concludes that during the reach phase, the receiver should look at the givers hand. This is also en par with findings in [2].
     * For the giver, both gaze control schemes 'receiver's face' and 'receiver's face -> receiver's hand' are both acceptable but it is summarized 
     * "that participants liked the transition gaze behaviors more than any of the fixed gaze behaviors" [1]. 
     * In [2], the authors state this even more clearly:
     * (1) "The participants preferred a gaze pattern in which the robot initially looks at their face and then transitions its gaze to their hand 
     * and then transitions its gaze back to look at their face again." [2]
     * (2) "The participants’ gaze preference did not change for changes in the object size, object fragility, or the user’s posture. However, 
     * the gaze preferences were also not statistically equivalent for different object size, object fragility, or the user’s posture" [2]
     */
    class Component :
        virtual public armarx::Component,
        virtual public armarx::view_selection::components::handover_target_provider::
            ComponentInterface,
        // , virtual public armarx::DebugObserverComponentPluginUser
        // , virtual public armarx::LightweightRemoteGuiComponentPluginUser
        virtual public armarx::ArVizComponentPluginUser,
        virtual public ObjectPoseClientPluginUser,
        // virtual public armarx::view_selection::components::person_target_provider::ComponentInterface,
        // virtual public armem::client::PluginUser
        // virtual public armem::client::MemoryListenerInterface
        virtual public armem::ListeningClientPluginUser
    {
    public:
        Component();

        /// @see armarx::ManagedIceObject::getDefaultName()
        std::string getDefaultName() const override;

        /// Get the component's default name.
        static std::string GetDefaultName();

        void
        start(const ::std::string& humanTrackingId,
              ::armarx::view_selection::components::handover_target_provider::RobotRole robotRole,
              const ::Ice::Current& = ::Ice::emptyCurrent) override;

        void stop(const ::Ice::Current& = ::Ice::emptyCurrent) override;

    protected:
        /// @see PropertyUser::createPropertyDefinitions()
        armarx::PropertyDefinitionsPtr createPropertyDefinitions() override;

        /// @see armarx::ManagedIceObject::onInitComponent()
        void onInitComponent() override;

        /// @see armarx::ManagedIceObject::onConnectComponent()
        void onConnectComponent() override;

        /// @see armarx::ManagedIceObject::onDisconnectComponent()
        void onDisconnectComponent() override;

        /// @see armarx::ManagedIceObject::onExitComponent()
        void onExitComponent() override;


        /* (Requires armarx::LightweightRemoteGuiComponentPluginUser.)
        /// This function should be called once in onConnect() or when you
        /// need to re-create the Remote GUI tab.
        void createRemoteGuiTab();

        /// After calling `RemoteGui_startRunningTask`, this function is
        /// called periodically in a separate thread. If you update variables,
        /// make sure to synchronize access to them.
        void RemoteGui_update() override;
        */

        void updateTargets();


    private:
        // Private methods go here.

        // Forward declare `Properties` if you used it before its defined.
        // struct Properties;

        /* (Requires the armarx::ArVizComponentPluginUser.)
        /// Draw some boxes in ArViz.
        void drawBoxes(const Properties& p, viz::Client& arviz);
        */

        armarx::SimplePeriodicTask<>::pointer_type task;


        static const std::string defaultName;

        // Private member variables go here.


        /// Properties shown in the Scenario GUI.
        struct Properties
        {
            std::string humanPoseProviderName = "OpenNIPointCloudProvider";

            // std::string objectPoseProviderDynamic = "sim_track";

            std::string robotName = "Armar6";

            float objectDetectionTimeout = 3.f;
            double trackTime = 5.0f;

            std::string targetLayerName = "targets";

            int updatePeriodMs = 100;
        };

        Properties properties_;
        /* Use a mutex if you access variables from different threads
         * (e.g. ice functions and RemoteGui_update()).
        std::mutex propertiesMutex;
        */


        /* (Requires the armarx::LightweightRemoteGuiComponentPluginUser.)
        /// Tab shown in the Remote GUI.
        struct RemoteGuiTab : armarx::RemoteGui::Client::Tab
        {
            armarx::RemoteGui::Client::LineEdit boxLayerName;
            armarx::RemoteGui::Client::IntSpinBox numBoxes;

            armarx::RemoteGui::Client::Button drawBoxes;
        };
        RemoteGuiTab tab;
        */


        /* (Requires the armarx::ArVizComponentPluginUser.)
         * When used from different threads, an ArViz client needs to be synchronized.
        /// Protects the arviz client inherited from the ArViz plugin.
        std::mutex arvizMutex;
        */

        VirtualRobot::RobotPtr robot_;

        armem::client::plugins::ReaderWriterPlugin<armem::robot_state::VirtualRobotReader>*
            virtualRobotReaderPlugin_ = nullptr;
        armem::client::plugins::ReaderWriterPlugin<armem::human::client::Reader>*
            humanPoseReaderPlugin_ = nullptr;

        client::ViewSelection viewSelection_;

        std::unique_ptr<target_provider::handover::HandoverTargetProvider> handoverTargetProvider_;
    };

} // namespace armarx::view_selection::components::handover_target_provider
