/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    navigation::ArmarXObjects::navigation_skill_provider
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @date       2023
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "Component.h"

#include <ArmarXCore/libraries/DecoupledSingleComponent/Decoupled.h>

#include <armarx/view_selection/skills/LookAt.h>
#include <armarx/view_selection/skills/LookAtArticulatedObjectFrame.h>
#include <armarx/view_selection/skills/LookAtHumanFace.h>
#include <armarx/view_selection/skills/LookAtHumanHand.h>
#include <armarx/view_selection/skills/LookAtObject.h>
#include <armarx/view_selection/skills/LookDirection.h>
#include <armarx/view_selection/skills/LookForObjects.h>
#include <armarx/view_selection/skills/LookForObjectsWithKinematicUnit.h>
#include <armarx/view_selection/skills/ResetGazeTargets.h>
#include <armarx/view_selection/skills/ScanLocation.h>
#include <armarx/view_selection/skills/ScanLocationsForObject.h>
#include <armarx/view_selection/skills/SetCustomGazeTarget.h>
#include <armarx/view_selection/skills/constants.h>
#include <armarx/view_selection/skills/skills.h>

namespace armarx::view_selection::components::view_selection_skill_provider
{
    Component::Component() : viewSelection(memoryNameSystem(), getName())
    {
        addPlugin(virtualRobotReaderPlugin);
        addPlugin(articulatedObjectReaderPlugin);
        addPlugin(objectInstanceReader);
    }

    const std::string Component::defaultName =
        view_selection::skills::constants::ViewSelectionSkillProviderName;

    armarx::PropertyDefinitionsPtr
    Component::createPropertyDefinitions()
    {
        armarx::PropertyDefinitionsPtr def =
            new armarx::ComponentPropertyDefinitions(getConfigIdentifier());

        def->component(kinematicUnit);

        // Publish to a topic (passing the TopicListenerPrx).
        // def->topic(myTopicListener);

        // Subscribe to a topic (passing the topic name).
        // def->topic<PlatformUnitListener>("MyTopic");

        // Use (and depend on) another component (passing the ComponentInterfacePrx).


        // Add a required property. (The component won't start without a value being set.)
        // def->required(properties.boxLayerName, "p.box.LayerName", "Name of the box layer in ArViz.");

        // Add an optionalproperty.

        auto optionalSubSkillID =
            [&def](armarx::skills::SkillID& skillID, const std::string& nameBase)
        {
            ARMARX_CHECK(skillID.providerId.has_value());
            def->optional(skillID.providerId->providerName, nameBase + "->providerName");
            def->optional(skillID.skillName, nameBase + ".skillName");
        };

        def->component(gazeScheduler);


        def->topic(this->textToSpeech, "TextToSpeech");

        def->optional(properties.humanPoseMemoryID.memoryName,
                      "p.HumanMemory.MemoryName",
                      "Memory name of the Human memory.");
        def->optional(properties.humanPoseMemoryID.coreSegmentName,
                      "p.HumanMemory.CoreSegmentName",
                      "Core segment name of the Human memory.");
        def->optional(properties.robotName, "p.RobotName", "Name of the robot, e.g., Armar6.");
        def->optional(
            properties.kinematicUnitGuiName, "p.KinematicUnit", "Name of the kinematic unit.");

        optionalSubSkillID(
            properties.scanLocationsForObject.subSkillIDs.lookAtArticulatedObjectFrame,
            "p.scanLocationsForObject.subSkillIDs.lookAtArticulatedObjectFrame");

        return def;
    }

    void
    Component::onInitComponent()
    {
        // Topics and properties defined above are automagically registered.

        // Keep debug observer data until calling `sendDebugObserverBatch()`.
        // (Requies the armarx::DebugObserverComponentPluginUser.)
        // setDebugObserverBatchModeEnabled(true);
    }

    void
    Component::onConnectComponent()
    {
        ARMARX_TRACE;
        // Do things after connecting to topics and components.
        viewSelection.connect();
        //ARMARX_CHECK_NOT_NULL(viewSelection);

        ARMARX_TRACE;
        ARMARX_CHECK_NOT_NULL(virtualRobotReaderPlugin);
        robot_ = virtualRobotReaderPlugin->get().getRobotWaiting(properties.robotName);
        ARMARX_CHECK_NOT_NULL(robot_);

        ARMARX_TRACE;
        ARMARX_CHECK_NOT_NULL(kinematicUnit);


        // {
        //     ARMARX_CHECK_NOT_NULL(virtualRobotReaderPlugin);
        //     ARMARX_CHECK_NOT_NULL(virtualRobotReaderPlugin);
        //     ARMARX_CHECK_NOT_NULL(roomsReaderPlugin);

        //     ARMARX_CHECK_NOT_NULL(skills_.guideHumanToRoom);

        //     skills::GuideHumanToRoom::Services srv{.iceNavigator = iceNavigator,
        //                                            .memoryNameSystem = memoryNameSystem(),
        //                                            .virtualRobotReader =
        //                                                virtualRobotReaderPlugin->get(),
        //                                            .costmapReader = costmapReaderPlugin->get(),
        //                                            .roomsReader = roomsReaderPlugin->get(),
        //                                            .arviz = arviz};
        //     skills_.guideHumanToRoom->connect(srv);
        // }


        ARMARX_TRACE;
        ARMARX_INFO << "add skill factories";
        // look in named directions
        {
            addSkillFactory<skills::LookUp>(
                skills::LookDirection::Context{.viewSelectionClient = viewSelection});
        }
        ARMARX_INFO << "add skill factories, line " << __LINE__;
        {
            addSkillFactory<skills::LookDown>(
                skills::LookDirection::Context{.viewSelectionClient = viewSelection});
        }
        ARMARX_INFO << "add skill factories, line " << __LINE__;
        {
            addSkillFactory<skills::LookDownstraight>(
                skills::LookDirection::Context{.viewSelectionClient = viewSelection});
        }
        ARMARX_INFO << "add skill factories, line " << __LINE__;
        {
            addSkillFactory<skills::LookLeft>(
                skills::LookDirection::Context{.viewSelectionClient = viewSelection});
        }
        ARMARX_INFO << "add skill factories, line " << __LINE__;
        {
            addSkillFactory<skills::LookRight>(
                skills::LookDirection::Context{.viewSelectionClient = viewSelection});
        }
        ARMARX_INFO << "add skill factories, line " << __LINE__;
        {
            addSkillFactory<skills::LookAhead>(
                skills::LookDirection::Context{.viewSelectionClient = viewSelection});
        }
        ARMARX_INFO << "add skill factories, line " << __LINE__;

        // custom gaze target
        {
            addSkillFactory<skills::SetCustomGazeTarget>(
                skills::SetCustomGazeTarget::Context{.viewSelectionClient = viewSelection});
        }
        ARMARX_INFO << "add skill factories, line " << __LINE__;

        // scan location
        {
            skills::ScanLocation::Services srv{.viewSelectionClient = viewSelection};
            addSkillFactory<skills::ScanLocation>(srv);
        }
        ARMARX_INFO << "add skill factories, line " << __LINE__;

        // look for objects
        {
            skills::LookForObjects::Services srv{
                .viewSelectionClient = viewSelection,
                .objectPoseClient = ObjectPoseClient::getClient(),
            };
            addSkillFactory<skills::LookForObjects>(srv);
        }
        ARMARX_INFO << "add skill factories, line " << __LINE__;
        {
            skills::LookForObjectsWithKinematicUnit::Services srv{
                .kinematicUnit = kinematicUnit,
                .objectPoseClient = getClient(),
                .textToSpeech = this->textToSpeech,
                .robotStateComponent = getRobotStateComponent()};
            addSkillFactory<skills::LookForObjectsWithKinematicUnit>(srv);
        }
        ARMARX_INFO << "add skill factories, line " << __LINE__;

        // look at
        {
            addSkillFactory<skills::LookAt>(
                skills::LookAt::Context{.viewSelectionClient = viewSelection});
        }

        ARMARX_INFO << "add skill factories, line " << __LINE__;

        // LookAtArticulatedObjectFrame
        {
            addSkillFactory<skills::LookAtArticulatedObjectFrame>(
                skills::LookAtArticulatedObjectFrame::Services{
                    .articulatedObjectReader = std::experimental::make_observer<
                        ::armarx::armem::articulated_object::Reader>(
                        &articulatedObjectReaderPlugin->get())});
        }
        ARMARX_INFO << "add skill factories, line " << __LINE__;

        {
            addSkillFactory<skills::LookAtHumanFace>(
                skills::LookAtHumanFace::Services{.memoryNameSystem = memoryNameSystem(),
                                                  .viewSelectionClient = viewSelection,
                                                  .humanPoseMemoryID = properties.humanPoseMemoryID,
                                                  .robot = robot_});
        }
        {
            addSkillFactory<skills::LookAtObject>(skills::LookAtObject::Services{
                .objectInstanceReader =
                    std::experimental::make_observer(&this->objectInstanceReader->get())});
        }
        ARMARX_INFO << "add skill factories, line " << __LINE__;
        ARMARX_INFO << "add skill factories, line " << __LINE__;
        {
            addSkillFactory<skills::LookAtHumanHand>(
                skills::LookAtHumanHand::Services{.memoryNameSystem = memoryNameSystem(),
                                                  .viewSelectionClient = viewSelection,
                                                  .humanPoseMemoryID = properties.humanPoseMemoryID,
                                                  .robot = robot_});
        }
        ARMARX_INFO << "add skill factories, line " << __LINE__;
        {
            skills::ScanLocationsForObject::Services services{
                .objectReader = this->objectInstanceReader->get()};
            addSkillFactory<skills::ScanLocationsForObject>(services,
                                                            properties.scanLocationsForObject);
        }

        ARMARX_INFO << "add skill factories, line " << __LINE__;
        {
            skills::ResetGazeTargets::Context services{.viewSelectionClient = gazeScheduler};
            addSkillFactory<skills::ResetGazeTargets>(services);
        }


        /* (Requies the armarx::DebugObserverComponentPluginUser.)
        // Use the debug observer to log data over time.
        // The data can be viewed in the ObserverView and the LivePlotter.
        // (Before starting any threads, we don't need to lock mutexes.)
        {
            setDebugObserverDatafield("numBoxes", properties.numBoxes);
            setDebugObserverDatafield("boxLayerName", properties.boxLayerName);
            sendDebugObserverBatch();
        }
        */

        /* (Requires the armarx::ArVizComponentPluginUser.)
        // Draw boxes in ArViz.
        // (Before starting any threads, we don't need to lock mutexes.)
        drawBoxes(properties, arviz);
        */

        /* (Requires the armarx::LightweightRemoteGuiComponentPluginUser.)
        // Setup the remote GUI.
        {
            createRemoteGuiTab();
            RemoteGui_startRunningTask();
        }
        */
        ARMARX_INFO << "added all skill factories";
    }

    void
    Component::onDisconnectComponent()
    {
    }

    void
    Component::onExitComponent()
    {
    }

    std::string
    Component::getDefaultName() const
    {
        return Component::defaultName;
    }

    std::string
    Component::GetDefaultName()
    {
        return Component::defaultName;
    }

    /* (Requires the armarx::LightweightRemoteGuiComponentPluginUser.)
    void
    Component::createRemoteGuiTab()
    {
        using namespace armarx::RemoteGui::Client;

        // Setup the widgets.

        tab.boxLayerName.setValue(properties.boxLayerName);

        tab.numBoxes.setValue(properties.numBoxes);
        tab.numBoxes.setRange(0, 100);

        tab.drawBoxes.setLabel("Draw Boxes");

        // Setup the layout.

        GridLayout grid;
        int row = 0;
        {
            grid.add(Label("Box Layer"), {row, 0}).add(tab.boxLayerName, {row, 1});
            ++row;

            grid.add(Label("Num Boxes"), {row, 0}).add(tab.numBoxes, {row, 1});
            ++row;

            grid.add(tab.drawBoxes, {row, 0}, {2, 1});
            ++row;
        }

        VBoxLayout root = {grid, VSpacer()};
        RemoteGui_createTab(getName(), root, &tab);
    }


    void
    Component::RemoteGui_update()
    {
        if (tab.boxLayerName.hasValueChanged() || tab.numBoxes.hasValueChanged())
        {
            std::scoped_lock lock(propertiesMutex);
            properties.boxLayerName = tab.boxLayerName.getValue();
            properties.numBoxes = tab.numBoxes.getValue();

            {
                setDebugObserverDatafield("numBoxes", properties.numBoxes);
                setDebugObserverDatafield("boxLayerName", properties.boxLayerName);
                sendDebugObserverBatch();
            }
        }
        if (tab.drawBoxes.wasClicked())
        {
            // Lock shared variables in methods running in seperate threads
            // and pass them to functions. This way, the called functions do
            // not need to think about locking.
            std::scoped_lock lock(propertiesMutex, arvizMutex);
            drawBoxes(properties, arviz);
        }
    }
    */


    /* (Requires the armarx::ArVizComponentPluginUser.)
    void
    Component::drawBoxes(const Component::Properties& p, viz::Client& arviz)
    {
        // Draw something in ArViz (requires the armarx::ArVizComponentPluginUser.
        // See the ArVizExample in RobotAPI for more examples.

        viz::Layer layer = arviz.layer(p.boxLayerName);
        for (int i = 0; i < p.numBoxes; ++i)
        {
            layer.add(viz::Box("box_" + std::to_string(i))
                      .position(Eigen::Vector3f(i * 100, 0, 0))
                      .size(20).color(simox::Color::blue()));
        }
        arviz.commit(layer);
    }
    */


    ARMARX_REGISTER_COMPONENT_EXECUTABLE(Component, Component::GetDefaultName());

} // namespace armarx::view_selection::components::view_selection_skill_provider
