/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     VS_CODE_USER_NAME ( VS_CODE_USER_EMAIL )
 * @date       2023
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <experimental/memory>

#include <Eigen/Core>

#include <VirtualRobot/VirtualRobot.h>

#include <ArmarXCore/core/services/tasks/TaskUtil.h>
#include <ArmarXCore/core/time.h>
#include <ArmarXCore/util/CPPUtility/TripleBuffer.h>

#include <RobotAPI/components/ArViz/Client/Client.h>
#include <RobotAPI/libraries/armem_robot_state/client/common/VirtualRobotReader.h>

#include <armarx/view_selection/gaze_targets/GazeTarget.h>

#include "ControllerHandlerInterface.h"

namespace armarx::view_selection::gaze_scheduler
{

    class Scheduler
    {
    public:
        struct InjectedServices
        {
            std::experimental::observer_ptr<armem::robot_state::VirtualRobotReader>
                virtualRobotReader;
            std::experimental::observer_ptr<viz::Client> arviz;

            std::experimental::observer_ptr<ControllerHandlerInterface> controllerHandler;
        };

        struct Params
        {
            std::string robotName;
            Eigen::Vector3f defaultTarget;

            bool visuEnabled;
            int targetVizAlpha;
        };

        Scheduler(const InjectedServices& srv, const Params& params);
        virtual ~Scheduler() = default;

        void handleTargetUpdate(const gaze_targets::GazeTarget& memoryTarget);
        void resetPriorityQueue();

        void submitToQueue(const gaze_targets::GazeTarget& target);

        // submit next target from queue based on priority and preemption attributes
        // WARNING: can only be safely called after acquiring the targetLock
        void scheduleNextTarget();

    private:
        VirtualRobot::RobotPtr robot_ = nullptr;

        //check whether the current target is out of time and reschedule
        void checkTargetDeadline();

        void submitControlTarget(const gaze_targets::GazeTarget& target);


        // an irreplaceable default target with low priority used if no other
        // targets are submitted
        void submitIdleTarget();


        void visualizeActiveTarget();

        void runPeriodically();


        const InjectedServices srv;
        const Params params;

        void init();


        // target managemant needs to be synchronized
        std::mutex targetMutex;
        std::multiset<gaze_targets::GazeTarget, std::greater<>> requestedTargets;

        TripleBuffer<std::optional<gaze_targets::GazeTarget>> currentTargetBuffer;
        armarx::DateTime currentDeadline = DateTime::Invalid();
        armarx::SimplePeriodicTask<>::pointer_type task;
    };


} // namespace armarx::view_selection::gaze_scheduler
