/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Niklas Arlt ( ujrwq at student dot kit dot edu )
 * @date       2023
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <RobotAPI/libraries/skills/provider/SimpleSpecializedSkill.h>

#include <armarx/view_selection/client/ViewSelection.h>
#include <armarx/view_selection/skills/aron/LookDirection.aron.generated.h>

namespace armarx::view_selection::skills
{
    namespace direction
    {
        const Eigen::Vector3f centerPosition{0., 1500., 1500.};
        const Eigen::Vector3f downstraightPosition{0., 1500., 500.};
        const Eigen::Vector3f DeltaUp{0., 0., 1000.};
        const Eigen::Vector3f deltaDown{0., 0., -1000.};
        const Eigen::Vector3f deltaLeft{-1000., 0., 0.};
        const Eigen::Vector3f deltaRight{1000., 0., 0.};

        /**
         * The namespace state contains the states of allowed directions.
         * There are 9 states in total:
         * Left+Up     Center+Up     Right + Up
         * Left+Center   Center    Right+Center
         * Left+Down  Center+Down   Right+Down
         */
        namespace state
        {
            using Type = std::uint8_t;
            constexpr std::uint8_t Center = 0;
            constexpr std::uint8_t Up = 0b1;
            constexpr std::uint8_t Down = 0b10;
            constexpr std::uint8_t Left = 0b1000;
            constexpr std::uint8_t Right = 0b10000;
            constexpr std::uint8_t Downstraight = 0b100000;

            using StateModFn = state::Type(state::Type);

            state::Type from(Eigen::Vector3f targetPosition);

            Eigen::Vector3f generateTargetPosition(state::Type state);

            state::Type center(state::Type previous);
            state::Type up(state::Type previous);
            state::Type down(state::Type previous);
            state::Type left(state::Type previous);
            state::Type right(state::Type previous);

            state::Type downstraight(state::Type previous);

        } // namespace state


    } // namespace direction

    /*
     * The LookDirection class implements all functions except DefaultSkillDescription().
     *
     * For each implemented direction there is a specialized subclass below.
     */

    class LookDirection :
        virtual public armarx::skills::SimpleSpecializedSkill<arondto::LookDirectionParams>
    {
    public:
        using Params = arondto::LookDirectionParams;
        using Base = ::armarx::skills::SimpleSpecializedSkill<Params>;

        struct Context
        {
            client::ViewSelection& viewSelectionClient;
            // RobotReader from RobotAPI? Would be another option to get current state.
        };

        LookDirection(const Context& services,
                      direction::state::StateModFn* stateModFn,
                      armarx::skills::SkillDescription skillDescription);

        void connect(const Context& srv);

    private:
        ::armarx::skills::Skill::InitResult init(const Base::SpecializedInitInput& in) override;
        ::armarx::skills::Skill::MainResult main(const Base::SpecializedMainInput& in) override;

        void onStopRequested() override;


    protected:
        static armarx::skills::SkillDescription
        generateDefaultSkillDescription(std::string name, std::string directionText);

    private:
        std::optional<Context> srv_;
        direction::state::StateModFn* stateModFn;

        // will be initialized in "init()"
        std::optional<gaze_targets::GazeTarget> gazeTarget;
    };

    /*
     * Specialized classes for each direction
     */

    class LookUp : public LookDirection
    {
    public:
        LookUp(const Context& services);

        static armarx::skills::SkillDescription DefaultSkillDescription();
    };

    class LookDown : public LookDirection
    {
    public:
        LookDown(const Context& services);

        static armarx::skills::SkillDescription DefaultSkillDescription();
    };

    /**
     * This direction forces to look down in the middle
    */
    class LookDownstraight : public LookDirection
    {
    public:
        LookDownstraight(const Context& services);

        static armarx::skills::SkillDescription DefaultSkillDescription();
    };

    class LookLeft : public LookDirection
    {
    public:
        LookLeft(const Context& services);

        static armarx::skills::SkillDescription DefaultSkillDescription();
    };

    class LookRight : public LookDirection
    {
    public:
        LookRight(const Context& services);

        static armarx::skills::SkillDescription DefaultSkillDescription();
    };

    class LookAhead : public LookDirection
    {
    public:
        LookAhead(const Context& services);

        static armarx::skills::SkillDescription DefaultSkillDescription();
    };
} // namespace armarx::view_selection::skills
