/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @date       2022
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "Human.h"

#include <optional>
#include <string>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <ArmarXCore/core/logging/Logging.h>

#include <VisionX/libraries/human/pose/model/k4a_bt_body_32.h>

namespace armarx::view_selection::targets
{

    Human::Human(const armem::human::HumanPose& human,
                 const Eigen::Vector3f& robotPosition,
                 const Params& params) :
        human_(human), robotPosition_(robotPosition), params_(params)
    {
    }

    std::optional<gaze_targets::GazeTarget>
    Human::asGazeTarget() const
    {
        ARMARX_CHECK(human_.humanTrackingId.has_value())
            << "A tracking id is required. Otherwise, the name will be ambiguous";

        // at the moment, we assume the azure kinect model
        const auto headKey = armarx::human::pose::model::k4a_bt_body_32::JointNames.to_name(
            armarx::human::pose::model::k4a_bt_body_32::Joints::Head);

        if (not(human_.keypoints.count(headKey) > 0))
        {
            ARMARX_VERBOSE << "head keypoint `" << headKey << "` not available";
            return std::nullopt;
        }

        const auto globalPosition = human_.keypoints.at(headKey).positionGlobal;
        if (not globalPosition.has_value())
        {
            ARMARX_VERBOSE << "global position not available";
            return std::nullopt;
        }

        gaze_targets::GazeTarget target;

        {
            target.name = "human_" + human_.humanTrackingId.value();
            target.position = armarx::FramedPosition(globalPosition->toEigen(), GlobalFrame, "");

            target.priority = gaze_targets::TargetPriority(gaze_targets::AttentionType::TaskDriven,
                                                           priority(globalPosition->toEigen()));
            target.duration = armarx::Duration::Seconds(1);
            target.creationTimestamp = armarx::Clock::Now();
        }

        return target;
    }

    double
    Human::priority(const Eigen::Vector3f& globalHeadPosition) const
    {
        const double distance = (globalHeadPosition - robotPosition_).norm();
        return 1 - (distance - params_.distanceMin) / (params_.distanceMax - params_.distanceMin);
    }


} // namespace armarx::view_selection::targets
